"""
    Command Parser with Long/Short options/arguments\n
    使用command.Command类创建指令模板，为模板添加长短选项（Options）、回调函数、别名、帮助信息等\n
    使用parser.CommandParser类创建对象，根据情况调用tryParse或parse解析指令
"""

from .__version__ import __version__
from .command import Command,OPT,CommandParser,ParseResult,CommandCore
from .event import Events
# Parser=CommandParser
# Core=CommandCore
#这样就可以 from LSParser import *


import os
import sys
import pytest

sys.path.insert(0,os.getcwd())

from LSparser.command import CommandCore


def teardown_function():
    #每个测试函数后执行
    CommandCore.cores.clear()
    CommandCore.last=None

def test_name():
    assert CommandCore.getLast().name==CommandCore.default
    assert CommandCore.last==CommandCore.default
    with CommandCore("admin") as core:
        assert core.name=="admin"
        assert CommandCore.last=="admin"
        with CommandCore("extra"):
            assert CommandCore.last=="extra"
        assert CommandCore.last=="admin"
    assert CommandCore.last==CommandCore.default
    for n in [CommandCore.default,"admin","extra"]:
        assert n in CommandCore.cores
    assert CommandCore.getCore("extra").name=="extra"
    assert CommandCore.getCore().name==CommandCore.default
    with pytest.raises(KeyError):
        CommandCore.getCore("void")

def test_dup():
    CommandCore.getLast()
    with pytest.raises(ValueError):
        CommandCore(CommandCore.default) # 创建重名指令中枢
    # print(err.value) # 指令管理器已存在

def test_prefix():
    core=CommandCore.getLast()
    assert "." in core.commandPrefix
    assert "。" in core.commandPrefix
    assert "-" in core.commandPrefix
    assert core.isMatchPrefix("!test")
    assert not core.isMatchPrefix("test")
    assert not core.isMatchPrefix("")


"""
事件管理
"""

from ..util.asyncs import callAllGen,asyncCallAllGen,emptyAsyncGen

class EventLists(list):

    @staticmethod
    def mergeLists(one,other):
        if not one:
            return other
        if isinstance(one,EventLists):
            one.merge(other)
        elif isinstance(other,EventLists):
            one=EventLists(one)
            one.merge(other)
        else:
            one+=other
        return one

    @property
    def subs(self):
        if self._subs is None:
            self._subs={}
        return self._subs

    def __init__(self,*args,**kw):
        self._subs=None
        super().__init__(*args,**kw)

    def __setattr__(self,name,value):
        if isinstance(value,(list,tuple)):
            self.subs[name]=EventLists(value)
        else:
            super().__setattr__(name,value)

    def __getattr__(self,name):
        subList=self.subs.get(name)
        if subList:
            return subList
        else:
            raise AttributeError

    def __delattr__(self, name):
        if name in self.subs:
            del self.subs[name]
        else:
            super().__delattr__(name)

    def __str__(self):
        return f"{super().__str__()}\nsubs:{self._subs}"

    def __bool__(self):
        return bool(len(self) or self._subs) #若列表或subs中有值，就为true

    def add(self,*callbacks):
        self+=callbacks

    def send(self,*args,**kw):
        return [*self.sendGen(*args,**kw)]

    def sendGen(self,*args,**kw):
        return callAllGen(self,*args,**kw)

    async def asyncSend(self,*args,**kw):
        return [r async for r in self.asyncSendGen(*args,**kw)]

    def asyncSendGen(self,*args,**kw):
        # async for r in asyncCallAllGen(self,*args,**kw):
        #     yield r
        return asyncCallAllGen(self,*args,**kw)

    def removes(self,*callbacks):
        for func in callbacks:
            try:
                self.remove(func)
            except:
                print(f"没有找到回调",func)
    
    def merge(self,other):
        if isinstance(other,(list,tuple)):
            self+=other
            if isinstance(other,EventLists):
                for k,v in other.subs.items():
                    self.addSub(k,*v)
        else:
            raise ValueError(f"无法将 {other} 纳入")

    def hasSub(self,name):
        return self._subs and name in self._subs #subs不为空且name在其中

    def getSub(self,name):
        return getattr(self,name,None)

    def addSub(self,name,*callbacks):
        subList=self.getSub(name)
        if subList:
            subList+=callbacks
        else:
            subList=[*callbacks]
        setattr(self,name,subList)

    def ensureSub(self,name):
        subList=self.getSub(name)
        if subList:
            return subList
        setattr(self,name,[])
        return self.getSub(name)

    ensure=ensureSub

    def sendSub(self,name,*args,**kw):
        return [*self.sendSubGen(name,*args,**kw)]

    def sendSubGen(self,name,*args,**kw):
        subList=self.getSub(name)
        if subList:
            yield from callAllGen(subList,*args,**kw)
        else:
            print(f"子事件 {name} 不存在！")

    async def asyncSendSub(self,name,*args,**kw):
        return [r async for r in self.asyncSendSubGen(name,*args,**kw)]

    def asyncSendSubGen(self,name,*args,**kw):
        subList=self.getSub(name)
        if subList:
            # async for r in asyncCallAllGen(subList,*args,**kw):
            #     yield r
            return asyncCallAllGen(subList,*args,**kw)
        else:
            print(f"子事件 {name} 不存在！")
        # async for r in emptyAsyncGen():
        #     yield r
        return emptyAsyncGen()

    def removeSub(self,name,*callbacks):
        subList:list=self.getSub(name)
        if subList:
            for func in callbacks:
                try:
                    subList.remove(func)
                except:
                    print(f"子事件 {name} 中没有找到回调",func)
        else:
            print(f"子事件 {name} 不存在！")
    
    def removeSubAll(self,name):
        if name in self.subs:
            self.subs.pop(name)
        else:
            print(f"子事件 {name} 不存在！")

EL=EventLists

class EventManager:
    
    # @classmethod
    # def callAll(cls,callList,*args,**kw):
        # return [*callAllGen(callList,*args,**kw)]
        # results=[]
        # for func in callList:
        #     rtmp=func(*args,**kw)
        #     results.append(rtmp)
        # return results
    
    def __init__(self):
        self.events={}

    def get(self,name):
        return self.events.get(name)

    def add(self,name,*callbacks):
        callList=self.get(name)
        if callList:
            EL.add(callList,*callbacks)
            # callList+=callbacks
        else:
            callList=[*callbacks] # EventLists(callbacks)
        self.events[name]=callList
    
    def ensure(self,name):
        callList=self.get(name)
        if callList:
            return callList
        callList=EventLists([])
        self.events[name]=callList
        return callList

    def send(self,name,*args,**kw):
        # callList=self.events.get(name)
        # if callList:
        #     return EventManager.callAll(callList,*args,**kw)
        # # print(f"事件 {name} 不存在！")
        # return []
        return [*self.sendGen(name,*args,**kw)]

    def sendGen(self,name,*args,**kw):
        callList=self.get(name)
        if callList:
            return EL.sendGen(callList,*args,**kw)
            # yield from callAllGen(callList,*args,**kw)
        return ()
    
    async def asyncSend(self,name,*args,**kw):
        return [r async for r in self.asyncSendGen(name,*args,**kw)]

    def asyncSendGen(self,name,*args,**kw):
        callList=self.get(name)
        if callList:
            return EL.asyncSendGen(callList,*args,**kw)
            # async for r in asyncCallAllGen(callList,*args,**kw):
            #     yield r
        # async for r in emptyAsyncGen():
        #     yield r
        return emptyAsyncGen()

    def remove(self,name,*callbacks):
        callList:list=self.get(name)
        if callList:
            print(f"尝试移除事件 {name} 的一些回调…")
            EL.removes(callList,*callbacks)
            # for func in callbacks:
            #     try:
            #         callList.remove(func)
            #     except:
            #         print(f"事件 {name} 中没有找到回调",func)
        else:
            print(f"事件 {name} 不存在！")
    
    def removeAll(self,name):
        if name in self.events:
            self.events.pop(name)
        else:
            print(f"事件 {name} 不存在！")

    def merge(self,name,otherName):
        callList=self.get(name)
        otherList=self.get(otherName)
        if otherList:
            callList=EL.mergeLists(callList,otherList)
            self.events[name]=callList
            self.removeAll(otherName)

    def hasSub(self,name,subName):
        callList=self.get(name)
        return callList and isinstance(callList,EventLists) and callList.hasSub(subName)

    def addSub(self,name,subName,*callbacks):
        callList=self.get(name)
        if callList:
            if not isinstance(callList,EventLists):
                callList=EventLists(callList)
        else:
            callList=EventLists()
        callList.addSub(subName,*callbacks)
        self.events[name]=callList
        

    def sendSub(self,name,subName,*args,**kw):
        # callList=self.get(name)
        # if callList and isinstance(callList,EventLists):
        #     return callList.sendSub(subName,*args,**kw)
        # # print(f"事件 {name} 的子事件 {subName} 不存在！")
        # return []
        return [*self.sendSubGen(name,subName,*args,**kw)]

    def sendSubGen(self,name,subName,*args,**kw):
        callList=self.get(name)
        if callList and isinstance(callList,EventLists):
            return callList.sendSubGen(subName,*args,**kw)
        return ()

    async def asyncSendSub(self,name,subName,*args,**kw):
        return [r async for r in self.asyncSendSubGen(name,subName,*args,**kw)]
    
    def asyncSendSubGen(self,name,subName,*args,**kw):
        callList=self.get(name)
        if callList and isinstance(callList,EventLists):
            # async for r in callList.asyncSendSubGen(subName,*args,**kw):
            #     yield r
            return callList.asyncSendSubGen(subName,*args,**kw)
        # async for r in emptyAsyncGen():
        #     yield r
        return emptyAsyncGen()

    def removeSub(self,name,subName,*callbacks):
        callList=self.get(name)
        if callList and isinstance(callList,EventLists):
            callList.removeSub(subName,*callbacks)
        else:
            print(f"事件 {name} 的子事件 {subName} 不存在！")

    def removeSubAll(self,name,subName):
        callList=self.get(name)
        if callList and isinstance(callList,EventLists):
            callList.removeSubAll(subName)
        else:
            print(f"事件 {name} 的子事件 {subName} 不存在！")


if __name__ == "__main__":

    # class CallableName:
    #     def __init__(self,name):
    #         self.name=name
    #     def __call__(self):
    #         print(self)
    #     def __str__(self):
    #         return self.name
    #     __repr__=__str__
    # Name=CallableName

    # A=Name("A");B=Name("B");C=Name("C");D=Name("D")
    # E=Name("E");F=Name("F");G=Name("G");H=Name("H")
    # I=Name("I");J=Name("J");K=Name("K");L=Name("L")
    # M=Name("M");N=Name("N");O=Name("O");P=Name("P")
    # Q=Name("Q");R=Name("R");S=Name("S");T=Name("T")
    # U=Name("U");V=Name("V");W=Name("W");X=Name("X")
    # Y=Name("Y");Z=Name("Z")

    # a=EventLists([A,B,C])
    # a.addSub("def",D,E,F)
    # print(a)
    # a+=[G,H,I]
    # print(a)
    # a.sendSub("h")
    # a.sendSub("def")
    # a.removeSub("def",E)
    # print(a)

    # em=EventManager()
    # em.addSub("b","jkl",J,K,L)
    # print(em.events["b"])
    # a.addSub("jkl",M)
    # print(a)
    # em.events["b"].merge(a)
    # print(em.events["b"])
    # em.removeSubAll("b","def")
    # print(em.events["b"])

    # c=EventLists()
    # print(c)
    # c.addSub("n",N)
    # print(bool(c))
    pass


            




import typing

def indent(texts:typing.Union[str,typing.Iterable],num=2,char=" ",fn:typing.Callable[...,str]=None):
    prefix=char*num
    if isinstance(texts,str):
        texts=texts.split("\n")
    if fn:
        return "\n".join([ f"{prefix}{fn(t)}" for t in texts ])
    else:
        return "\n".join([ f"{prefix}{t}" for t in texts ])


from ..command import CommandCore

class CommandEvents:
    """
        指令事件相关操作
    """

    @staticmethod
    def getEventName(name):
        """
            指令事件名统一为 cmd-指令名
        """
        return f"cmd-{name}"

    Error="error" # error子事件（报错回调）的名称

    def __init__(self,name,core:CommandCore):
        """
            name 为指令名\n
            core 为指令所在的中枢对象
        """
        self.name=name
        self.core=core
        self.EM=self.core.EM
        self.eventName=CommandEvents.getEventName(self.name)

    def onExecute(self,*callbacks):
        """
            添加执行回调
        """
        self.EM.add(self.eventName,*callbacks)
        print(f"为 {self.name} 指令添加了执行回调")
    
    def onError(self,*callbacks):
        """
            添加报错回调
        """
        self.EM.addSub(self.eventName,CommandEvents.Error,*callbacks)
        print(f"为 {self.name} 指令添加了报错回调")
    
    def merge(self,otherName):
        """
            将其他名称的指令对应的事件纳入当前指令的事件中
        """
        otherEventName=CommandEvents.getEventName(otherName)
        if otherEventName in self.EM.events:
            print(f"将 {otherName} 指令的回调纳入 {self.name} 指令")
            self.EM.merge(self.eventName,otherEventName)

    def send(self,name,pr=None,*args,**kw):
        """
            触发回调
        """
        if not pr is None:
            result=[]
            for r in self.EM.sendGen(name,*args,**kw):
                pr.output.append(r)
                result.append(r)
            return result
        return self.EM.send(name,*args,**kw)

    async def asyncSend(self,name,pr=None,*args,**kw):
        """
            异步触发回调
        """
        if not pr is None:
            result=[]
            async for r in self.EM.asyncSendGen(name,*args,**kw):
                pr.output.append(r)
                result.append(r)
            return result
        return await self.EM.asyncSend(name,*args,**kw)

    def sendExecute(self,pr=None,*args,**kw):
        """
            触发执行回调\n
                pr 解析结果对象，不为 None 则回调结果会传入其 output 中
        """
        return self.send(self.eventName,pr,*args,**kw)

    async def asyncSendExecute(self,pr=None,*args,**kw):
        """
            异步触发执行回调\n
                pr 解析结果对象，不为 None 则回调结果会传入其 output 中
        """
        return await self.asyncSend(self.eventName,pr,*args,**kw)

    def sendSub(self,name,subName,pr=None,*args,**kw):
        """
            触发子事件回调
        """
        if not pr is None:
            result=[]
            for r in self.EM.sendSubGen(name,subName,*args,**kw):
                pr.output.append(r)
                result.append(r)
            return result
        return self.EM.sendSub(name,subName,*args,**kw)

    async def asyncSendSub(self,name,subName,pr=None,*args,**kw):
        """
            异步触发子事件回调
        """
        if not pr is None:
            result=[]
            async for r in self.EM.asyncSendSubGen(name,subName,*args,**kw):
                pr.output.append(r)
                result.append(r)
            return result
        return await self.EM.asyncSend(name,*args,**kw)

    def sendError(self,pr=None,*args,**kw):
        """
            触发报错回调\n
                pr 解析结果对象，不为 None 则回调结果会传入其 output 中
        """
        return self.sendSub(self.eventName,CommandEvents.Error,pr,*args,**kw)

    async def asyncSendError(self,pr=None,*args,**kw):
        """
            异步触发报错回调\n
                pr 解析结果对象，不为 None 则回调结果会传入其 output 中
        """
        return await self.asyncSendSub(self.eventName,CommandEvents.Error,pr,*args,**kw)

# 这个在py3.9以下用不了……
# 只能使用 event.py 中的替代方案
class CommandEventsWrapper:
    """
        给指令添加事件用的装饰器
    """

    def __init__(self,name,coreName=None):
        """
            name 指令名\n
            coreName 指令中枢名，默认为 None
        """
        core=CommandCore.getCore(coreName)
        cmd=core.cmds.get(name)
        if cmd:
            self.events=cmd.events
        else:
            self.events=CommandEvents(name,core)
    
    def __call__(self,func):
        """
            等同于添加执行回调
        """
        return self.execute(func)

    def execute(self,func):
        """
            添加执行回调
        """
        self.events.onExecute(func)
        return func

    def error(self,func):
        """
            添加报错回调
        """
        self.events.onError(func)
        return func


import os
import sys
import pytest

sys.path.insert(0,os.getcwd())

from LSparser.command import CommandCore,OPT,Option,CommandHelper,OptType
from LSparser import Events,Command,CommandParser,ParseResult

def teardown_function():
    #每个测试函数后执行
    CommandCore.cores.pop("main")
    CommandCore.last=None

def test_basic():
    with CommandCore("main") as core:
        c=Command("test")
        assert c.name=="test"
        assert c.core.name==core.name
        assert c.typelist==core.commandPrefix
        assert c.name in c.core.cmds
        c.names("tes","est")
        assert "tes" in c.nameset
        assert "est" in c.nameset
        with pytest.raises(ValueError):
            c.names("test")
        with pytest.raises(ValueError):
            Command("test")
        with pytest.raises(ValueError):
            Command("")
        c=Command("cmd").opt("-s1",OPT.Try).opt("-s2",OPT.Must).opt("--l",OPT.Must)
        assert "-s1" in c.shortOpts
        assert "-s2" in c.shortOpts
        assert "--l" in c.longOpts
        assert OPT.M==OPT.Must
        assert c.shortOpts["-s1"].hasValue==OPT.T
        assert c.longOpts["--l"] in c.Optset
        c=Command("!solo")
        assert len(c.typelist)==1 and "!" in c.typelist
        assert not core.isMatchPrefix("$multi")
        c=Command("multi",types=["。","$"])
        assert len(c.typelist)==2 and "。" in c.typelist and "$" in c.typelist
        assert core.isMatchPrefix("$multi")

def test_opt():
    with CommandCore("main"):
        with pytest.raises(ValueError):
            Option([])
        with pytest.raises(ValueError):
            Option("")
        assert Option.getOptType("--s") == OptType.Long
        assert Option.getOptType("-!d") == OptType.Short
        assert Option.getOptType("not") == OptType.Not
        assert Option(["-s","--c","-!d"]).name=="s"
        assert Option(["--c","-!d","-s"]).name=="c"
        assert Option("-e",OPT.M).hasValue==OPT.Must
        assert "-ads" in Option("ads").names

def test_event():
    with CommandCore("main") as core:
        cp=CommandParser()

        def p(result:ParseResult):
            print(result)
            return "success"
        Command("test").onExecute(p)
        pr=cp.tryParse(".test")
        assert pr.isChecked()
        assert pr.isCommand()
        assert pr.isDefinedCommand()
        assert not pr.isWrongType()
        assert not pr.hasOpt()
        assert pr.command=="test"
        assert not pr.params
        assert pr.paramStr==""
        assert pr.output[0] == "success"

        Command("deco")
        @Events.onCmd("deco")
        def _(result):
            return "deco"
        assert cp.tryParse(".deco").output[0] == "deco"

        @Events.onCmd("duck")
        def _(result):
            return result.command

        @Events.onCmd("duk")
        def _(result):
            return result.command

        c=Command("duck").names("dack","duk","daaa")
        assert len(core.EM.events[c.events.eventName])==2
        assert not core.EM.events.get("cmd-duk")
        out=cp.tryParse(".duk").output
        assert out[0]=="duk" and out[1]=="duk"

        @Events.onCmd("lag")
        def _(result):
            return result.command
        c=Command("lag").names("laag","lagg","log")
        assert len(core.EM.events[c.events.eventName])==1
        @Events.onCmd("laag")
        def _(result):
            return result.command
        assert len(core.EM.events[c.events.eventName])==2
        assert not core.EM.events.get("cmd-laag")
        @Events.onCmd("lag")
        def _(result):
            return result.command
        assert len(core.EM.events[c.events.eventName])==3
        assert cp.tryParse(".log").realCommand == "lag"
        
        @Events.onCmd("error")
        def _(result):
            return [][5]
        
        @Events.onCmd.error("error")
        def _(result,err:Exception):
            assert str(err)=="list index out of range"
            return err

        c=Command("error")
        pr=cp.tryParse("!error")
        assert len(pr.output)==1
        assert isinstance(pr.output[0],IndexError)

        @Events.onCmd("no-error")
        def _(result):
            return [][5]
        
        c=Command("no-error")
        with pytest.raises(IndexError):
            pr=cp.tryParse("!no-error")

        @Events.onCmd.error("before")
        def beforeError(result):
            return result.command
        @Events.onCmd("before")
        def _(result):
            return result.command
        @Events.onCmd.error("bf")
        def bfError(result):
            return result.command
        c=Command("before").names("bf")
        assert core.EM.hasSub(c.events.eventName,"error")
        assert len(core.EM.events[c.events.eventName])==1
        assert len(core.EM.events[c.events.eventName].subs["error"])==2
        @Events.onCmd.error("bf")
        def _(result):
            return result.command
        assert len(core.EM.events[c.events.eventName].subs["error"])==3

        @Events.onCmd("bf")
        def _(result):
            return result.command
        assert len(core.EM.events[c.events.eventName])==2

        @Events.onCmd.error("af")
        def _(result):
            return result.command
        @Events.onCmd("af")
        def _(result):
            return result.command
        c=Command("after").names("af")
        assert core.EM.hasSub(c.events.eventName,"error")
        assert len(core.EM.events[c.events.eventName])==1
        assert len(core.EM.events[c.events.eventName].subs["error"])==1

        outputLen=0
        @Events.onCmd("ultra-error")
        @Events.onCmd("ultra-error")
        @Events.onCmd("ultra-error")
        def onUltra(result):
            nonlocal outputLen
            # print(result.output)
            assert len(result.output)==outputLen
            outputLen+=1
            return result.command
        
        @Events.onCmd("ultra-error")
        def _(result):
            return [][5]
        
        Events.onCmd("ultra-error")(onUltra)
        Events.onCmd("ultra-error")(onUltra)
        Events.onCmd("ultra-error")(onUltra)

        @Events.onCmd.error("ultra-error")
        @Events.onCmd.error("ultra-error")
        @Events.onCmd.error("ultra-error")
        def _(result,err):
            nonlocal outputLen
            # print(result.output)
            assert len(result.output)==outputLen
            outputLen+=1
            return err

        c=Command("ultra-error")
        assert len(core.EM.events[c.events.eventName])==7
        assert len(core.EM.events[c.events.eventName].subs["error"])==3
        pr=cp.tryParse("!ultra-error")
        assert len(pr.output)==outputLen==6
        assert pr.output[0]==pr.command
        assert isinstance(pr.output[-1],IndexError)


def test_parse():
    with CommandCore("main"):
        cp=CommandParser()
        Command("cmd")
        assert cp.tryParse(".cmd M W WmW").params == ["M","W","WmW"]
        assert cp.tryParse(".cmd -r -s W --t").paramStr == "-r -s W --t"
        Command("shortN").opt("-s",OPT.Not)
        assert not cp.tryParse(".shortN").args
        assert not cp.tryParse(".shortN --s").args
        assert not cp.tryParse(".shortN -t").args
        assert cp.tryParse(".shortN -s").args["s"]
        assert cp.tryParse(".shortN -s abc").args["s"] is True
        Command("shortM").opt("-s",OPT.Must).opt("-t")
        assert cp.tryParse(".shortM -s").args["s"] #只有选项，没提供值，故为True
        assert cp.tryParse(".shortM -s abc").args["s"] == "abc"
        assert cp.tryParse(".shortM -s abc -s cde fg -s h").args["s"] == "h" #后面的覆盖前面的
        pr=cp.tryParse(".shortM -s -t")
        assert pr.args["s"] == "-t" and not pr.args.get("t") #-s必须有值
        assert pr.getByType("s") == "-t"
        assert cp.tryParse(".shortM -t -s").args["t"]
        Command("shortT").opt("-s",OPT.Try).opt("-t")
        assert cp.tryParse(".shortT -s").args["s"] #只有选项，没提供值，故为True
        pr=cp.tryParse(".shortT -s -t")
        assert pr.args["s"] is True and pr.args["t"] is True #因为有-t 故 -s 放弃值，只为True
        assert pr.getByType("s",T=bool) is True
        assert pr.getByType("t","Not str!") == "Not str!"
        assert pr.getByType("NO","NO!!!") == "NO!!!"
        assert pr.splitHead("sho") == "rtT"
        assert pr.splitHead("ssh") == ""
        Command("longN").opt("--s",OPT.Not)
        assert not cp.tryParse(".longN -s").args
        assert cp.tryParse(".longN --s").args["s"] is True
        Command("longM").opt("--s",OPT.Must).opt("-t",OPT.Must)
        assert cp.tryParse(".longM --s").args["s"] == [] #没提供值的时候是空列表
        assert cp.tryParse(".longM --s a b c").args["s"] == ["a","b","c"]
        assert cp.tryParse(".longM --s a b -t c").args["s"] == ["a","b"]
        assert cp.tryParse(".longM --s -t c").args["s"] == [] #被 -t 挡住的时候会是空列表
        assert not cp.tryParse(".longM -t --s c").args.get("s")
        Command("longT").opt("--s",OPT.Try).opt("-t")
        assert cp.tryParse(".longT --s").args["s"] is True #没提供值的时候是True
        assert cp.tryParse(".longT --s a b c").args["s"] == ["a","b","c"]
        assert cp.tryParse(".longT --s -t c").args["s"] is True #被 -t 挡住的时候会是True

def test_parse_events():
    with CommandCore("main") as core:
        cp=CommandParser()

        beforeTimes=0
        afterTimes=0
        @Events.onBeforeParse
        def before(pr:ParseResult,cp:CommandParser):
            nonlocal beforeTimes
            pr.data["brain"]="void"
            assert pr.parser==cp
            assert pr.isChecked()
            beforeTimes+=1

        @Events.onAfterParse
        def after(pr:ParseResult,cp:CommandParser):
            nonlocal afterTimes
            assert pr.data["brain"]=="void"
            assert pr.parser==cp
            afterTimes+=1

        @Events.onNotCmd()
        def notCmd(pr:ParseResult,cp:CommandParser):
            assert not pr.raw or pr.raw=="hh" or (not pr.command and pr.type=="。")

        @Events.onUndefinedCmd("main")
        def undefined(pr:ParseResult,cp:CommandParser):
            assert pr.command=="nono"
            pr.opt("-s").opt("--l",OPT.M).opt("-c",OPT.T).parse()
            assert pr["s"] is True
            assert pr["l"]==["yes","yes"]
            assert pr["c"]=="wow"
            assert pr.paramStr=="!"

        @Events.onWrongCmdType
        def wrong(pr:ParseResult,cp:CommandParser):
            assert pr.command=="uno"
            assert pr.type=="!"
            assert not "s" in pr
            pr.type="$"
            pr.parse()
            assert pr["s"]=="uno!"

        Command("cmd").opt("-s").opt("-c",OPT.M)
        Command("uno",types=["$","#"]).opt("-s",OPT.M)

        cp.tryParse("hh")
        cp.tryParse("")
        cp.tryParse("。")

        cp.tryParse(".cmd -s -c md")
        cp.tryParse(".nono -s --l yes yes -c wow !")
        cp.tryParse("!uno -s uno!")
        assert beforeTimes==6
        assert afterTimes==6

        @Events.onExecuteError
        def onError(pr:ParseResult, cp:CommandParser, err:Exception) -> bool:
            assert isinstance(err,IndexError)
        
        @Events.onCmd("raise")
        @Events.onCmd("error")
        def error(pr:ParseResult):
            [][863]
        
        Command("error")
        with pytest.raises(IndexError):
            cp.tryParse(".error")
        
        @Events.onCmd.error("error")
        def errorError(pr:ParseResult,err:Exception):
            assert isinstance(err,IndexError)
        
        assert cp.tryParse(".error").output==[None]

        Command("raise")
        with pytest.raises(IndexError):
            cp.tryParse(".raise")

        @Events.onExecuteError
        def handleError(pr:ParseResult, cp:CommandParser, err:Exception) -> bool:
            assert isinstance(err,IndexError)
            return True
        
        assert cp.tryParse(".raise").output==[]
              

def test_data():
    class StringContaier:
        extra="more than a string!"

        def __init__(self,s):
            self.s=s

        def __str__(self) -> str:
            return self.s

    with CommandCore("main"):
        cp=CommandParser()
        cp.data["one"]="shot"
        Command("cmd")

        @Events.onBeforeParse
        def before(pr:ParseResult, cp:CommandParser):
            assert pr.parserData["one"]=="shot"
            assert not "two" in pr.data
        
        @Events.onCmd("cmd")
        def cmd(pr:ParseResult):
            assert pr.raw.extra=="more than a string!"
            pr.data["two"]="kick"
        
        @Events.onAfterParse
        def after(pr:ParseResult, cp:CommandParser):
            assert pr.parserData["one"]=="shot"
            assert pr.data["two"]=="kick"

        cp.tryParse( StringContaier(".cmd dmc") )

        Command("extra")
        @Events.onBeforeParse
        def _(pr:ParseResult, cp:CommandParser):
            ex1,ex2=pr.dataArgs
            kw=pr.dataKW
            ex3=kw["ex3"]
            ex4=kw["ex4"]
            assert ex1==42
            assert ex2=="TTT"
            assert ex3==...
            assert ex4==(True,False)

        @Events.onCmd("extra")
        def _(pr:ParseResult):
            pr.data["two"]="kick"
            return pr.dataArgs,pr.dataKW

        out=cp.tryParse(".extra",42,"TTT",ex3=...,ex4=(True,False)).output[0]
        assert out[0]==(42,"TTT")
        assert out[1]=={"ex3":...,"ex4":(True,False)}

def test_help():
    with CommandCore("main"):
        c=Command("cmd").names("CMD","Command")\
            .opt(["-s","--straight"],OPT.Must,"val  Straight!")\
            .opt(["-t","--tight"],OPT.Try,"[val]  Tight!")\
            .opt(["--l","--long"],OPT.Not," Long!")\
            .help("这是cmd",
            """
                cmd具有强大的功能
                其中之一便是显示帮助
                非常多的帮助
            """,
            ".cmd C M D",
            ".cmd -s smile",
            ".cmd -t T?",
            ".cmd --l"
            )
        assert c.frontHelp=="这是cmd\ncmd具有强大的功能\n其中之一便是显示帮助\n非常多的帮助"
        assert c.backHelp==".cmd C M D\n.cmd -s smile\n.cmd -t T?\n.cmd --l"
        helper=CommandHelper()
        # helper=CommandHelper("./help/default")
        assert helper.getHelp([".cmd"]).strip() == c.getHelp()
        # print(c.getHelp())
        # helper.lineLimit=2
        # print(helper.getHelp([".cmd"],1))
        # print(helper.getHelp([".cmd"],2))
        # print(helper.getHelp([".cmd"],10))
        # print(helper.getHelp([".cmd"],-5))
        # helper.generateHelp([".cmd"])
        # helper.generateMainHelp(startText="指令详情\n",endText="\n========")
        # print(helper.getHelp([".cmd","--l","-t","-s"]))
        



#coding:utf-8
"""
    解析指令
"""
#from .util import *
from . import CommandCore,BaseCommand,Command,Option,OPT,OptType
from ..event import EventNames

from enum import IntEnum

import asyncio

class CommandState(IntEnum):
    Unknown=-1          #未解析
    NotCommand=0        #不是指令
    Command=1           #是指令（但不一定用指令模板定义过）
    DefinedCommand=2    #是指令模板定义过的指令
    WrongType=3         #是指令模板定义过的指令（但解析时发现指令前缀不匹配）

class ParseResult:
    """
        指令解析结果
    """

    def __init__(self,parser:"CommandParser"=None):
        self.command=""
        self.type=""
        self.params=[]
        self.args={}
        self.raw="" #完整的被解析字符串
        self.state=CommandState.Unknown #通过CommandState枚举记录指令解析状态
        self._cons=[]
        self._cmd:BaseCommand=None
        self.output=[]
        self._data=None #用来携带一些额外的数据
        self.parser=parser
        # if self.parser:
        #     self.data["parser"]=self.parser._data # 从 parser 中继承 _data 的引用

    def __contains__(self,key):
        return key in self.args

    def __getitem__(self,key):
        return self.args.get(key) #访问args的简易写法

    def __str__(self):
        result=f"{self.raw}\n"
        result+=f"type: {self.type or '未知'}\n"
        result+=f"command: {self.command or '未知'}\n"
        result+=f"params: {self.params}\n"
        if self.hasOpt:
            for k,v in self.args.items():
                result+=f"{k}: {v}\n"
        return result.strip() #基本输出的是 self 的内容，只不过格式漂亮一点

    __repr__=__str__

    def __bool__(self):
        return self.isCommand()

    @property
    def data(self):
        """
            可携带的一些额外数据
        """
        if not self._data:
            self._data={}
        return self._data

    @data.setter
    def data(self,val):
        self._data=val

    @property
    def dataArgs(self):
        """
            tryParse 方法中额外的 *args
        """
        return self.data.get("args",())

    @dataArgs.setter
    def dataArgs(self,val):
        if val:
            self.data["args"]=val

    @property
    def dataKW(self):
        """
            tryParse 方法中额外的 **kw
        """
        return self.data.get("kw",{})

    @dataKW.setter
    def dataKW(self,val):
        if val:
            self.data["kw"]=val

    @property
    def parserData(self):
        """
            解析器可能携带的一些额外数据
        """
        if self.parser:
            return self.parser.data

    def hasOpt(self):
        """
            指令是否有选项
        """
        if self._cmd:
            return len(self._cmd.Optset)!=0
        return False

    def isChecked(self):
        """
            指令是否已被初步检查（确定是否为指令、指令类型等）
        """
        return self.state != CommandState.Unknown

    def isCommand(self):
        """
            当前解析的内容是否是指令
        """
        return self.state.value > CommandState.NotCommand.value

    def isDefinedCommand(self):
        """
            当前解析的内容是否是指令模板定义过的指令
        """
        return self.state.value>=CommandState.DefinedCommand.value

    def isWrongType(self):
        """
            解析结果是否和指令模板定义的types不匹配
        """
        return self.state==CommandState.WrongType

    @property
    def paramStr(self):
        """
            得到用" "连接的指令参数（params）
        """
        return " ".join(self.params)

    @property
    def realCommand(self):
        """
            得到真正的指令名（而不是指令的别名）
        """
        if self._cmd:
            return self._cmd.name
        return self.command

    def getByType(self,arg,default=None,T=str):
        """
            得到解析后的指定选项，附加类型限制\n
                arg 选项的名称，不带选项前缀\n
                default 默认值，如果没得到选项则使用该值 默认为None\n
                T 要获得的类型，如果存在选项但不是指定类型的，也只返回默认值 默认为str
        """
        r=self.args.get(arg,default)
        if isinstance(r,T):
            return r
        return default

    def getToType(self,arg,default=None,T=str):
        """
            得到解析后的指定选项，附加类型限制\n
                arg 选项的名称，不带选项前缀\n
                default 默认值，如果没得到选项则使用该值 默认为None\n
                T 要获得的类型，如果存在选项但不是指定类型的，则尝试转换，转换失败时返回默认值 默认为str
        """
        r=self.args.get(arg,default)
        if isinstance(r,T):
            return r
        elif r:
            try:
                r=T(r)
                return r
            except:
                pass
        return default

    def splitHead(self,head):
        """
            若给定文本（head）在指令首部，则返回切除该文本后的字符串\n
            如解析到指令"rd"，传入"r"，则返回"d"
        """
        if self.command.startswith(head):
            return self.command[len(head):]
        return ""

    def opt(self,names,hasValue=OPT.Not):
        """
            在解析前为指令添加选项，常用于为没有用指令模板定义过的指令补充选项
        """
        self._cmd.opt(names,hasValue)
        return self

    def parse(self):
        """
            让解析器解析指令\n
            需要在为指令添加好所有选项后调用\n
        """
        if self.parser:
            return self.parser.parse(self)
        raise ValueError("没找到解析器！可能本解析结果不是通过解析器产生的…")

class CommandParser:
    '''
        指令解析类\n
        例: .cmd param1 param2 -s1 -s2 s2val --l lval1 lval2 lval3\n
    '''
    def __init__(self,coreName=None):
        """
            coreName 指令中枢名称，默认为 None，即最后创建的中枢\n
        """
        self._data=None
        self.result=ParseResult(parser=self)
        self.core=CommandCore.getCore(coreName)
        self.core.tryRefresh() #如果都要解析了，core还需要刷新，那就刷新一下

    @property
    def data(self):
        if not self._data:
            self._data={}
        return self._data

    @data.setter
    def data(self,val):
        self._data=val

    def opt(self,names,hasValue=OPT.Not):
        """
            在解析前为指令添加选项，常用于为没有用指令模板定义过的指令补充选项
        """
        self.result.opt(names,hasValue)
        return self

    def getCommand(self,t) -> ParseResult:
        """
            判断指定文本是否为指令，并更新当前解析状态\n
                t 文本\n
        """
        result=ParseResult(parser=self)
        result.raw=t
        if not isinstance(t,str):
            t=str(t)
        if t=="":
            result.state=CommandState.NotCommand
            self.result=result
            return result
        if self.core.isMatchPrefix(t): #至少被识别为一个指令
            result._cons=t.strip().split()
            result.type=result._cons[0][0]
            result.command=result._cons[0][1:]
            if result.command=="": #不允许空指令
                result.state=CommandState.NotCommand
                self.result=result
                return result
            result.state=CommandState.Command
            cmd:BaseCommand=self.core.cmds.get(result.command)
            
            if cmd: #定义的指令
                result._cmd=cmd
                result.state=CommandState.DefinedCommand
                if not result.type in cmd.typelist:
                    result.state=CommandState.WrongType
            else: #未定义指令，给它一个临时的模板
                result._cmd=BaseCommand(result._cons[0],coreName=self.core.name)
        else: #不是一个指令
            result.state=CommandState.NotCommand
        self.result=result
        return result
    
    def parse(self,pr:ParseResult=None):
        """
            解析指令\n
            需要在为指令添加好所有选项后调用\n
                pr 供解析用的 ParseResult 对象
        """
        if pr is None:
            pr=self.result
        con=pr._cons[1:]
        if not pr.hasOpt():
            pr.params=con
            self.result=pr
            return pr
        while con:
            optType=Option.getOptType(con[0])
            if optType==OptType.Not:
                pr.params.append(con[0])
                con=con[1:]
            else:
                matched=False
                if optType==OptType.Long:
                    tarOpt:Option=pr._cmd.longOpts.get(con[0])
                    if tarOpt:
                        matched=True
                        optText=tarOpt.name
                        _,pr.args[optText],con=self.getLong(con,tarOpt.hasValue)
                else: #OptType.Short
                    tarOpt:Option=pr._cmd.shortOpts.get(con[0])
                    if tarOpt:
                        matched=True
                        optText=tarOpt.name
                        _,pr.args[optText],con=self.getShort(con,tarOpt.hasValue)
                if not matched:
                    pr.params.append(con[0])
                    con=con[1:]
        self.result=pr
        return pr

    def getLong(self,con,hasValue):
        """
            解析中得到一个长选项
        """
        arg=con[0]
        con=con[1:]
        if hasValue!=OPT.Not:
            val=[]
            while con and Option.getOptType(con[0])==OptType.Not:
                val.append(con[0])
                con=con[1:]
            if hasValue==OPT.Try and len(val)==0:
                val=True
        else:
            val=True
        return arg,val,con

    def getShort(self,con,hasValue):
        """
            解析中得到一个短选项
        """
        arg=con[0]
        con=con[1:]
        if hasValue==OPT.Not or (not con) or (hasValue==OPT.Try and Option.getOptType(con[0])!=OptType.Not ):
            val=True
        else:
            val=con[0]
            con=con[1:]
        return arg,val,con


    def tryParse(self,t,*args,**kw):
        """
            尝试解析给定文本\n
            解析指令、调用回调函数、返回解析结果对象\n
                t 文本或可以转换成文本的对象\n
                剩余参数将作为解析结果对象的额外数据\n
        """
        pr=self.getCommand(t)
        pr.dataArgs=args
        pr.dataKW=kw
        self.core.EM.send(EventNames.BeforeParse,pr,self)
        if pr and pr.isDefinedCommand():
            pr=self.parse(pr)
            cmd:Command=pr._cmd
            try:
                cmd.events.sendExecute(pr,pr) #第二个 pr 会传入回调函数
                #第一个 pr 使得回调结果存入 output 中
            except Exception as err:
                cmdErrorHandle=cmd.events.sendError(pr,pr,err) #第二个 pr 和 err 会传入回调函数
                errorHandle=self.core.EM.send(EventNames.ExecuteError,pr,self,err)
                if not cmdErrorHandle and not any(errorHandle):
                    raise err #没处理的话还是要抛出错误的
        elif not pr.isCommand():
        #各种事件
            self.core.EM.send(EventNames.NotCmd,pr,self)
        elif not pr.isDefinedCommand():
            self.core.EM.send(EventNames.UndefinedCmd,pr,self)
        elif pr.isWrongType():
            self.core.EM.send(EventNames.WrongCmdType,pr,self)
            # self.sendEvents(pr)
        self.core.EM.send(EventNames.AfterParse,pr,self)
        return pr

    # def sendEvents(self,result:ParseResult):
    #     if not result.isCommand():
    #         self.core.EM.send(EventNames.NotCmd,result,self)
    #     elif not result.isDefinedCommand():
    #         self.core.EM.send(EventNames.UndefinedCmd,result,self)
    #     elif result.isWrongType():
    #         self.core.EM.send(EventNames.WrongCmdType,result,self)

    async def asyncTryParse(self,t,*args,**kw):
        """
            尝试解析给定文本\n
            解析指令、异步调用回调函数、返回解析结果对象\n
                t 文本或可以转换成文本的对象\n
                剩余参数将作为解析结果对象的额外数据\n
        """
        pr=self.getCommand(t)
        pr.dataArgs=args
        pr.dataKW=kw
        await self.core.EM.asyncSend(EventNames.BeforeParse,pr,self)
        if pr and pr.isDefinedCommand():
            pr=self.parse(pr)
            cmd:Command=pr._cmd
            try:
                await cmd.events.asyncSendExecute(pr,pr) #第二个 pr 会传入回调函数
                #第一个 pr 使得回调结果存入 output 中
            except Exception as err:
                cmdErrTsk=asyncio.create_task( cmd.events.asyncSendError(pr,pr,err) ) #第二个 pr 和 err 会传入回调函数
                parserErrTsk=asyncio.create_task( self.core.EM.asyncSend(EventNames.ExecuteError,pr,self,err) )
                cmdErrorHandle=await cmdErrTsk
                errorHandle=await parserErrTsk #暂时不知道这样搞好不好
                if not cmdErrorHandle and not any(errorHandle):
                    raise err #没处理的话还是要抛出错误的
        elif not pr.isCommand():
        #各种事件
            await self.core.EM.asyncSend(EventNames.NotCmd,pr,self)
        elif not pr.isDefinedCommand():
            await self.core.EM.asyncSend(EventNames.UndefinedCmd,pr,self)
        elif pr.isWrongType():
            await self.core.EM.asyncSend(EventNames.WrongCmdType,pr,self)
            # self.sendEvents(pr)
        await self.core.EM.asyncSend(EventNames.AfterParse,pr,self)
        return pr


if __name__=="__main__":
    cp=CommandParser()
    #cmd="！login 1313123 -a -z -s 888 --c 12 13 -x"
    cmd=".cmd param1 param2 -s1 -s2 s2val --long lval1 lval2 lval3"
    #cmd=".send aaa -box2"
    if cp.getCommand(cmd):
        #print(cp.separateCommand("lo"))
        #print(cp.separateCommand("logi"))
        #cp.opt(["-z","-a"],2).opt("--c",1).parse()
        #cp.opt("-s1",2).opt("-s2",1).opt(["--l","--long"],1).parse()
        #cp.opt(["-box","-b"],0).opt("-user",0).opt(["-qq","-group"],0).parse()
        print(cp.result)
    else:
        print("不是命令")
            
            
        
        



"""
    创建指令模板
"""

from os import terminal_size
from . import CommandCore
from ..event import CommandEvents
from ..util.str import indent

from enum import IntEnum

class OptVal(IntEnum):
    """
        用于Option的取值判定\n
            Not / N=0 无需值\\
            Must / M=1 需要值\\
            Try / T=2 尽可能有值\\
    """
    Not=0
    Must=1
    Try=2
    N=0
    M=1
    T=2

OPT=OptVal

class OptType(IntEnum):
    """
        用于Option的类型判定\n
            Not=0 非Opt\\
            Short=1 短\\
            Long=2 长\\
    """
    Not=0
    Short=1
    Long=2

class BaseCommand:
    """
        指令模板基类，包含一些指令模板的基本逻辑
    """
    def __init__(self,name,types:list=None,coreName=None):
        core=CommandCore.getCore(coreName)
        self.core=core
        if types is None:
            types=self.core.commandPrefix #使用中枢的默认前缀
        if self.core.isMatchPrefix(name):
            self.name=name[1:]
            self.typelist=[name[0]]
        else:
            self.name=name
            self.typelist=[]
            self.types(*types) #用 types 方法能更新中枢的潜在前缀
        if not self.name:
            raise ValueError("指令名不能为空")
        self.nameset=set()
        self.nameset.add(self.name)
        self.Optset=set()
        self.longOpts={}
        self.shortOpts={}

    def __str__(self):
        return f"Command - {self.name}\nOpts - {self.Optset}"

    def addLSOpt(self,name,opt,optType):
        """
            optType   
                OptType.Short：短\\
                OptType.Long：长
        """
        if optType==OptType.Short:
            opts=self.shortOpts
        elif optType==OptType.Long:
            opts=self.longOpts
        else:
            raise ValueError(f"{self.name} 指令的 {name} 选项的选项类型有误")
        if name in opts:
            raise ValueError(f"{self.name} 指令的 {name} 选项已存在")
        opts[name]=opt

    def types(self,*types):
        """
            为指令模板添加更多 type（指令前缀）\n
            支持链式调用
        """
        if self.typelist is self.core.commandPrefix:
            self.typelist=self.core.commandPrefix.copy()
        self.typelist+=types
        self.core.potentialPrefix.update(*types)
        return self
    
    def names(self,*names):
        """
            为指令模板添加更多名称（别名）\n
            支持链式调用
        """
        for n in names:
            self.nameset.add(n)
            # self.core.addCommand(self,name=n)
            # self.events.merge(n)
        return self

    def opt(self,names,hasValue=OptVal.Not):
        """
            为指令模板添加选项\n
                names 选项名，字符串/列表/元组，形如 "-a" 或 ["-a","-A","--aAa"]\n
                hasValue 取值判定\n
                    OPT.Not=0 不需要值\\
                    OPT.Must=1 需要值\\
                    OPT.Try=2 尽可能有值\n
            支持链式调用
        """
        Option(names,hasValue,None,self)
        return self
    

class Command(BaseCommand):
    """
        指令模板类
    """

    def __init__(self,name,types=None,coreName=None):
        """
            name 指令名，诸如 .test 或 test\n
            types 指令前缀列表。可用前缀会根据name决定，或默认使用指令中枢中的前缀\n
            coreName 要添加指令的指令中枢名称，默认为 None，即最后创建的中枢\n
        """

        super().__init__(name,types,coreName)
        
        self.help() #初始化时，默认不显示帮助，需要手动添加
        #添加到core
        self.core.addCommand(self)
        self.events:CommandEvents=CommandEvents(self.name,self.core)

    def __str__(self):
        return self.getHelp()

    __repr__=__str__

    def names(self,*names):
        """
            为指令模板添加更多名称（别名）\n
            支持链式调用
        """
        for n in names:
            self.nameset.add(n)
            self.core.addCommand(self,name=n) #将别名也添加到core
            self.events.merge(n) #如果定义了别名的事件，将它们划为指令的事件
        return self

    def opt(self,names,hasValue=OptVal.Not,help=None):
        """
            为指令模板添加选项\n
                names 选项名，字符串/列表/元组，形如 "-a" 或 ["-a","-A","--aAa"]\n
                hasValue 取值判定\n
                    OPT.Not=0 不需要值\\
                    OPT.Must=1 需要值\\
                    OPT.Try=2 尽可能有值\n
                help 选项的帮助文本\n
            支持链式调用
        """
        Option(names,hasValue,help,self)
        return self

    def getHelp(self):
        """
            得到指令的完整帮助文本
        """
        result=self.headHelp
        if self.showHelp:
            # .cmd/.CMD
            #   简介
            #   详细介绍
            #   -a 选项说明
            #   -b/--b 选项说明
            #   --c 选项说明
            # 例：
            #   例1
            #   例2
            #   ……
            # front="\n".join([ f"  {h}" for h in self.frontHelp.split("\n") ])
            front=indent(self.frontHelp)
            result+=f"\n{front}"
            if self.Optset:
                # optHelp="\n".join([ f"  {str(o)}" for o in self.Optset ])
                optHelp=indent(Option.sortOpts(*self.Optset),fn=str)
                result+=f"\n{optHelp}"
            if self.helpCase:
                result+="\n例："
                cases=indent(self.helpCase,fn=str)
                # cases="\n".join([ f"  {c}" for c in self.helpCase ])
                result+=f"\n{cases}"
        else: # .cmd/.CMD -a -b/--b --c 
            result+=" "
            result+=" ".join([ "/".join(o.names) for o in self.Optset ])
        return result.strip()

    def help(self,short:str=None,detail:str=None,*cases):
        """
            为指令模板添加帮助文本\n
                short 指令的简短介绍\n
                detail 指令的详细描述\n
                *cases 指令的例子合集\n
            不提供任何参数，则视为不显示帮助文本\n
            支持链式调用
        """
        self.shortHelp=""
        self.detailHelp=""
        self.helpCase=tuple()
        if not (short or detail or cases):
            self.showHelp=False
        else:
            self.showHelp=True
            if short:
                self.shortHelp=short.strip()
            if detail:
                self.detailHelp="\n".join([ l.strip() for l in detail.splitlines() if l.strip() ]) 
            if cases:
                self.helpCase=cases
        return self

    @property
    def headHelp(self):
        """
            帮助文本-头部，即指令名部分
        """
        text=""
        prefix=self.typelist[0]
        text+=f"{prefix}{self.name}"
        names=[*self.nameset]
        names.sort()
        for n in names:
            if n!=self.name:
                text+=f"/{prefix}{n}"
        return text

    @property
    def frontHelp(self):
        """
            帮助文本-前段
        """
        text=""
        if self.shortHelp:
            text+=self.shortHelp
        if self.detailHelp:
            text+="\n"
            text+=self.detailHelp
        return text.strip()

    @property
    def backHelp(self):
        """
            帮助文本-后段
        """
        return "\n".join(self.helpCase)

    def onExecute(self,*funcs):
        """
            为指令模板添加更多执行回调\n
            返回值统一存入解析结果的 `output` 列表中\n
            回调函数：`(result:ParseResult) -> Any`\n
            支持链式调用
        """
        self.events.onExecute(*funcs)
        return self

    callback=onExecute

    def onError(self,*funcs):
        """
            为指令模板添加更多报错回调\n
            返回值代替执行回调的结果存入解析结果的 `output` 列表中\n
            回调函数：`(result:ParseResult, err:Exception) -> Any`\n
            支持链式调用
        """
        self.events.onError(*funcs)
        return self

class Option:
    """
        选项类\n
        选项前缀见 Option.shortPrefix 和 Option.longPrefix
    """
    shortPrefix="-"
    longPrefix="--"

    @classmethod
    def getOptType(cls,text):
        """
            返回OptType
        """
        if text.startswith(cls.longPrefix):
            return OptType.Long
        elif text.startswith(cls.shortPrefix):
            return OptType.Short
        else:
            return OptType.Not
    
    @classmethod
    def sortOpts(cls,*opts):
        """
            将给定的各选项按名称、选项类型排序，短选项在长选项前
        """
        optNames=[*opts]
        optNames.sort(key=lambda o: o.names[0])
        if optNames and cls.getOptType(optNames[0].names[0])==OptType.Long: # 排序后可能长选项排在短选项前面
            shortIdx=len(optNames)
            for i,o in enumerate(optNames): # 找到列表中第一个短选项
                if cls.getOptType(o.names[0])==OptType.Short:
                    shortIdx=i
                    break
            optNames=optNames[shortIdx:]+optNames[:shortIdx] # 短选项在前，长选项在后
        return optNames


    def __init__(self,names,hasValue=OptVal.Not,help=None,cmd:Command=None):
        """
            详见 Command.opt\n
                cmd 要添加选项的指令模板
        """
        if not names:
            raise ValueError("选项不能为空")
        if isinstance(names,str):
            names=[names]
        for i,n in enumerate(names):
            optType=Option.getOptType(n)
            if optType==OptType.Not:
                names[i]=Option.shortPrefix + n
                optType=OptType.Short
            if cmd:
                cmd.addLSOpt(names[i],self,optType)
            if i==0:
                self.name=names[0].lstrip(Option.longPrefix if optType==OptType.Long else Option.shortPrefix)
        self.names=names
        self.hasValue=hasValue
        self.help=help
        if cmd:
            cmd.Optset.add(self)

    def __str__(self):
        """
            -a/-b/--c 这里是帮助
        """
        front="/".join(self.names)
        # if self.hasValue==OptVal.Not:
        #     mid=" True "
        # elif self.hasValue==OptVal.Must:
        #     mid=" Text "
        # else:
        #     mid=" True/Text "
        back=""
        if self.help:
            back=self.help
        # return front+mid+back
        return f"{front} {back}"

    __repr__=__str__

if __name__ == "__main__":
    c=Command("test").opt("-v",0,"测试选项").opt(["--r","-e"],1,"选项选项").help("这是测试命令")
    print( CommandCore.getLast() )
    print(c)
    print(OptVal.Must)
    print(OptVal.Must.value)
    print(OptVal.Must==1) #True


"""
    辅助查看、生成指令的帮助信息
"""
from . import CommandCore,Command,Option
from ..util.str import indent

import os
from io import StringIO

class CommandHelper:
    """
        帮助管理器
    """
    
    lineLimit=10
    
    # HintNotFound="是没见过的帮助呢"
    NotFound=""
    
    defaultFile="general.txt"
    defaultHidden="__hidden.txt"

    def __init__(self,root="./help",core=None):
        """
            root 帮助文本文件所在的根目录\n
            core 可以为指令中枢对象，也可以为其名称，默认为 None，即最后创建的中枢\n
        """
        if isinstance(core,CommandCore):
            self.core=core
        else:
            self.core=CommandCore.getCore(core)
        self.rootPath=root

    def getHelp(self,target:list=[],page=1):
        """
            按页查看指令的帮助信息\n
                target 目标指令，可以为字符串或列表，内容为指令名或包含其下的选项名\n
                page 页数，从 1 开始，默认为 1\n
                    每页行数由 cls.lineLimit 确定
        """
        cmd,target=self.tryFindCmd(target) # 检查、更新target，并尝试寻找指令模板
        filePath=self.getHelpFilePath(target) 
        if filePath: # 如果找到了文件，直接使用文件内容
            content=self.NotFound
            with open(filePath,"r",encoding="utf-8") as f:
                content=self.getPageContent(f,page)
            return content
        helpText=self.getHelpInCmd(cmd,target) # 否则去指令对象中找
        if helpText:
            return self.getPageContent(StringIO(helpText),page)
        return self.NotFound

    def tryFindCmd(self,target:list):
        """
            检查 target 并尝试得到指令模板
        """
        if isinstance(target,str): # str -> list
            target=target.split()
        cmd:Command=None
        if target:
            if self.core.isMatchPrefix(target[0]): #除去指令前缀
                target[0]=target[0][1:]
            cmd=self.core.cmds.get(target[0])
            if cmd:
                target[0]=cmd.name #真正的指令名
        return cmd,target

    def getPageContent(self,lines,page=1):
        """
            得到指定页的内容
        """
        page=max(page,1)
        currentPage=1
        lineNum=0
        pageContent=None
        current=[]
        last=[]
        for line in lines:
            current.append(line)
            lineNum+=1
            if lineNum==self.lineLimit: #该翻页了
                if currentPage==page:
                    pageContent=current
                currentPage+=1
                last=current
                current=[]
                lineNum=0
        totalPage=currentPage #总页数
        if totalPage>1 and len(current)<=self.lineLimit//2: #最后一页行数太少
            for line in current: #最后一页合并到倒数第二页
                last.append(line)
            totalPage-=1
            current=last
        if not pageContent:
            pageContent=current
        if pageContent and not pageContent[-1].endswith("\n"): #加入换行，让页码能显示在正确的位置
            pageContent[-1]+="\n"
        page=min(totalPage,page)
        if totalPage>1:
            pageContent.append(f"==={page}/{totalPage}===") #添加页码
        return "".join(pageContent)

    def getHelpFilePath(self,target:list):
        """
            尝试得到指令帮助文本文件的路径
        """
        path=os.path.join(self.rootPath,*target)
        filePath=path+".txt" # 先找 name.txt
        if not os.path.exists(filePath):
            filePath=os.path.join(path,self.defaultFile) # 再找 name/general.txt
            if not os.path.exists(filePath):
                return None
        return filePath

    def getHelpInCmd(self,cmd:Command,target):
        """
            尝试从指令模板中得到帮助信息
        """
        if not (cmd and cmd.showHelp and target):
            return None
        if len(target)==1:
            return cmd.getHelp()
        else: #得到给定选项的帮助
            opts=[]
            for n in target[1:]:
                opt=cmd.shortOpts.get(n) or cmd.longOpts.get(n)
                if opt:
                    opts.append(opt)
            if opts:
                result=cmd.headHelp
                optHelp=indent(Option.sortOpts(*opts),fn=str)
                result+=f"\n{optHelp}"
                return result
        return None

    def generateHelp(self,target:list=[]):
        """
            生成指令的帮助文本文件\n
                target 目标指令，可以为字符串或列表，内容为指令名或包含其下的选项名\n
        """
        cmd,target=self.tryFindCmd(target) # 检查、更新target，并尝试寻找指令模板
        helpText=self.getHelpInCmd(cmd,target) # 去指令对象中找帮助文本
        if helpText:
            filePath=self.getHelpFilePath(target) #优先覆盖已存在的
            if not filePath:
                filePath=os.path.join(self.rootPath,*target,self.defaultFile) #默认是 name/general.txt
            dir=os.path.dirname(filePath)
            self.ensureDir(dir)
            with open(filePath,"w",encoding="utf-8") as f:
                f.write(helpText)
            print("生成了帮助文本：",filePath)

    def generateMainHelp(self,startText="",endText=""): # 生成根目录的帮助文本，简单列出所有的指令
        """
            生成根目录的帮助文本文件，即简单的指令列表\n
                startText 在指令列表前附加额外内容\n
                endText 在指令列表后附加额外内容\n
        """
        self.ensureDir(self.rootPath)
        helps={}
        hiddens=self.getHidden(self.rootPath)
        for n in os.listdir(self.rootPath):
            name,_=os.path.splitext(n)
            if name in helps or (hiddens and name in hiddens) or n==self.defaultFile or n==self.defaultHidden:
                # hiddens 中的名称全部忽略
                # defaultFile 和 defaultHidden 也忽略
                continue
            filePath=os.path.join(self.rootPath,n) #非目录直接拿文件
            if os.path.isdir(filePath):
                filePath=os.path.join(filePath,self.defaultFile) #拿目录里的 general.txt
            if os.path.exists(filePath):
                helps[name]="  ".join(self.getFileShort(filePath))
        for cn in self.core.cmdlist:
            cmd:Command=self.core.cmds[cn]
            if cmd.name in helps or (hiddens and cmd.name in hiddens):
                # hiddens 中的名称全部忽略
                continue
            if cmd.showHelp: # 只添加显示帮助信息的指令
                helps[cmd.name]=self.getCmdShort(cmd)
        content=""
        for n,v in helps.items():
            if not self.core.isMatchPrefix(v): # 对于一些不以指令名开头的，加上指令名
                cmd=self.core.cmds.get(n)
                if cmd:
                    v=f"{cmd.typelist[0]}{n}  {v}"
                else:
                    # v=f"{self.core.commandPrefix[0]}{n}  {v}" # 对于尚未定义的指令，可能指令前缀会出错
                    v=f"{n}  {v}" # 先不添加指令前缀
            content+=f"{v}\n"
        if startText:
            content=f"{startText}\n{content}"
        if endText:
            content=f"{content}{endText}"
        mainPath=os.path.join(self.rootPath,self.defaultFile)
        with open(mainPath,"w",encoding="utf-8") as f:
            f.write(content)
        print("生成了主要帮助文本：",mainPath)
                

    def getFileShort(self,path,lineLimit=2):
        """
            得到帮助文本文件的前几行，作为指令简介
        """
        lines=[]
        count=0
        with open(path,"r",encoding="utf-8") as f:
            for line in f:
                lines.append(line.strip())
                count+=1
                if count==lineLimit:
                    break
        return lines

    def getCmdShort(self,cmd:Command):
        """
            从指令模板得到指令简介
        """
        if cmd.shortHelp:
            return f"{cmd.headHelp}  {cmd.shortHelp}"
        else:
            return cmd.headHelp

    def getHidden(self,dirPath):
        """
            得到目录中标记隐藏指令的文件
        """
        hiddenPath=os.path.join(dirPath,self.defaultHidden)
        if not os.path.exists(hiddenPath):
            return None
        hiddens=set()
        with open(hiddenPath,"r",encoding="utf-8") as f:
            for line in f:
                hiddens.add(line.strip())
        return hiddens
    
    def ensureDir(self,dir):
        """
            保证目录存在
        """
        if not os.path.exists(dir):
            os.makedirs(dir,exist_ok=True)

        


                

            


            

        





import os
import sys
import pytest
import asyncio

sys.path.insert(0,os.getcwd())

# from LSparser.command import CommandCore,OPT,Option,CommandHelper,OptType
from LSparser import Events,Command,CommandCore,CommandParser,OPT,ParseResult

def teardown_function():
    #每个测试函数后执行
    CommandCore.cores.pop("main")
    CommandCore.last=None

async def parse_main():
    with CommandCore("main") as core:
        cp=CommandParser()

        c=Command("single")
        c.opt("-s1",OPT.Try).opt("-s2",OPT.Must).opt("--s3",OPT.Must)

        @Events.onCmd("single")
        async def _(pr:ParseResult):
            assert not pr.dataArgs
            assert pr.dataKW["X"]==5 and pr.dataKW["Y"]==3

        await cp.asyncTryParse(".single only one -s1 + -s2 + --s3 - - + ",X=5,Y=3)

        inOrder=[]
        outOrder=[]
        @Events.onCmd("task1")
        @Events.onCmd("task2")
        @Events.onCmd("task3")
        async def _(pr:ParseResult):
            inOrder.append(pr.command)
            await asyncio.sleep(int(pr.command[-1])/100)
            outOrder.append(pr.command)
        Command("task1")
        Command("task2")
        Command("task3")

        await asyncio.gather( 
            cp.asyncTryParse(".task2"),
            cp.asyncTryParse(".task1"),
            cp.asyncTryParse(".task3")
        )
        assert inOrder==["task2","task1","task3"]
        assert outOrder==["task1","task2","task3"]

        sleepTick=0
        @Events.onCmd("taskM")
        @Events.onCmd("taskM")
        @Events.onCmd("taskM")
        @Events.onCmd("taskM")
        @Events.onCmd("taskM")
        @Events.onCmd("taskM")
        async def _(pr:ParseResult):
            nonlocal sleepTick
            local=sleepTick
            sleepTick+=1
            await asyncio.sleep(sleepTick/20)
            assert len(pr.output)==local
            return len(pr.output)

        Command("taskM")
        pr=await cp.asyncTryParse(".taskM")
        assert len(pr.output)==sleepTick==6
        assert pr.output==[0,1,2,3,4,5]
        
        actual=[]
        Command("mix")
        @Events.onCmd("mix")
        async def ac(pr:ParseResult):
            actual.append("a")
            return "a"
        @Events.onCmd("mix")
        def sc(pr:ParseResult):
            actual.append("s")
            return "s"

        Events.onCmd("mix")(ac)
        Events.onCmd("mix")(ac)
        Events.onCmd("mix")(sc)

        pr=await cp.asyncTryParse(".mix")
        assert actual==["a","a","a","s","s"]
        assert pr.output==["a","s","a","a","s"]

        Command("error")
        
        @Events.onCmd("error")
        async def _(pr:ParseResult):
            pr.notAttribute

        @Events.onCmd.error("error")
        async def aErr(pr:ParseResult, err:Exception):
            assert isinstance(err,AttributeError)
            assert not pr.output
            return "solved"

        @Events.onCmd.error("error")
        def sErr(pr:ParseResult, err:Exception):
            assert isinstance(err,AttributeError)
            assert len(pr.output)==1
            return "solved"

        Events.onCmd.error("error")(aErr) #异步先执行，结果开始传入outpu时已执行结束，故output长都是0

        @Events.onExecuteError
        async def _(pr:ParseResult, cp:CommandParser, err:Exception):
            assert not pr.output
            while not pr.output:
                await asyncio.sleep(0)
            assert pr.output

        await cp.asyncTryParse("!error")

def test_parse():
    asyncio.run(parse_main())

def test_sync():
    with CommandCore("main") as core:
        cp=CommandParser()
        
        actual=[]
        Command("mix")
        @Events.onCmd("mix")
        async def ac(pr:ParseResult):
            actual.append("a")
            return "a"
        @Events.onCmd("mix")
        def sc(pr:ParseResult):
            actual.append("s")
            return "s"

        Events.onCmd("mix")(ac)
        Events.onCmd("mix")(ac)
        Events.onCmd("mix")(sc)
        pr=cp.tryParse(".mix")
        assert actual==["a","a","a","s","s"]
        assert pr.output==["a","s","a","a","s"]


"""
    若干种字符串相似度
"""

import math

def getStrVector(tarstr,cdict,clen=None):
    """
        将字符串向量化，向量的每一项对应字符集中一种字符的频数，字符集、每种字符在向量中对应的下标由cdict提供
    """
    if not clen:
        clen=len(cdict.keys())
    vec=[0]*clen
    for c in tarstr:
        vec[cdict[c]]+=1
        #vec[cdict[c]]=1
    return vec

def cosSimilarity(str1,str2):
    """
        余弦相似度，将字符串向量化后计算
    """
    cdict={}
    currentidx=0
    for c in str1+str2:
        idx=cdict.get(c,None)
        if idx is None:
            cdict[c]=currentidx
            currentidx+=1
    vec1=getStrVector(str1,cdict,clen=currentidx)
    vec2=getStrVector(str2,cdict,clen=currentidx)
    dot=0
    sum1=0
    sum2=0
    for i in range(len(vec1)):
        sum1+=vec1[i]**2
        sum2+=vec2[i]**2
        dot+=vec1[i]*vec2[i]
    if dot==0:
        return 0.0
    return dot/( math.sqrt(sum1)*math.sqrt(sum2) )

def IoU(str1,str2):
    """
        交并比，将字符串转化为字符集合进行计算
    """
    set1=set(str1)
    set2=set(str2)
    ul=len(set1.union(set2))
    if ul==0:
        return 0.0
    return len(set1.intersection(set2))/ul

def diceSimilarity(str1,str2):
    """
        dice相似度，IoU的变种
    """
    set1=set(str1)
    set2=set(str2)
    ul=len(set1)+len(set2)
    if ul==0:
        return 0.0
    return 2*len(set1.intersection(set2))/ul

def editSimilarity(str1,str2):
    """
        编辑相似度，由编辑距离转化而来
    """
    len1=len(str1)
    len2=len(str2)
    lev=levenshtein(str1,str2,len1,len2)
    return 1-lev/max(len1,len2)

def levenshtein(str1,str2,len1=None,len2=None):
    """
        编辑距离，返回从一个字符串变成另一个，需要的最少编辑（插入、删除、替换字符）次数
    """
    if not len1:
        len1=len(str1)
    if not len2:
        len2=len(str2)
    #if len1==0:
    #    return len2
    #if len2==0:
    #    return len1
    dislist=[ [i for i in range(len2+1)] ]
    for i in range(1,len1+1):
        dislist.append( [0]*(len2+1) )
        dislist[i][0]=i
    for i in range(1,len1+1):
        for j in range(1,len2+1):
            temp=0
            if str1[i-1]==str2[j-1]:
                temp=0
            else:
                temp=1
            dislist[i][j]=min(dislist[i-1][j-1]+temp,dislist[i][j-1]+1,dislist[i-1][j]+1)
    #for d in dislist:
    #    print(d)
    return dislist[len1][len2]
    
def hammingSimilarity(str1,str2):
    """
        汉明相似度，表示相同长度字符串，对应位置上字符相同的程度
    """
    len1=len(str1)
    len2=len(str2)
    if len1!=len2:
        return 0.0
    if len1==0:
        return 1.0
    same=0
    for i in range(len1):
        if str1[i]==str2[i]:
            same+=1
    return same/len1
    


if __name__ == "__main__":
    #print( cosSimilarity("apple","app") )
    #print( cosSimilarity("abcd","dcba") )
    #print( cosSimilarity("非空","") )
    
    #print( IoU("aa","bb"),diceSimilarity("aa","bb") )
    #print( IoU("aa","ab"),diceSimilarity("aa","ab") )
    #print( IoU("aa",""),diceSimilarity("aa","") )
    #print( IoU("abcd","dcba"),diceSimilarity("abcd","dcba") )
    #print( IoU("abcd","aacbcdc"),diceSimilarity("abcd","aacbcdc") )
    #print( IoU("app","apple"),diceSimilarity("app","apple") )

    #print( editSimilarity("apple","app") )
    #print( editSimilarity("一二三四五","12345") )
    #print( editSimilarity("abcd","dcba") )
    #print( editSimilarity("非空","") )
    #print( editSimilarity("拥有效果","效果的") )
    #print( editSimilarity("效果的","拥有效果") )
    #print( editSimilarity("abcd","baaccd") )
    #print( editSimilarity("一样","一样") )

    #print( hammingSimilarity("一样","一样") )
    #print( hammingSimilarity("吃肉大过年的","大过年的吃肉") )
    #print( hammingSimilarity("一三一四二","二四一四三") )
    #print( hammingSimilarity("","") )

    print("全完了")




import asyncio
import typing

def asSync(coro:typing.Coroutine):
    """
        同步执行协程
    """
    newLoop=False
    try:
        loop=asyncio.get_event_loop()
    except RuntimeError:
        loop=asyncio.new_event_loop()
        newLoop=True
    result=loop.run_until_complete(coro)
    if newLoop:
        loop.close()
    return result

def gatherAsSync(*coros):
    newLoop=False
    try:
        loop=asyncio.get_event_loop()
    except RuntimeError:
        loop=asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        newLoop=True
    result=loop.run_until_complete( asyncio.gather( *coros ) )
    if newLoop:
        asyncio.set_event_loop(None)
        loop.close()
    return result

def callAllGen(callList,*args,**kw):
    """
        生成器，执行列表中所有回调，逐一返回结果
    """
    _,_,asyncs,asyncIdx=splitCorosWithIdx(callList)
    aLen=0
    aIdx=0
    if asyncs:
        aResults=gatherAsSync( *[coro(*args,**kw) for coro in asyncs] )
        aLen=len(aResults)
    for i,func in enumerate(callList):
        if aIdx<aLen and i==asyncIdx[aIdx]:
            yield aResults[aIdx]
            aIdx+=1
        else:
            yield func(*args,**kw)

async def asyncCallAllGen(callList,*args,**kw):
    """
        异步生成器，执行列表中所有回调，逐一返回结果
    """
    _,_,asyncs,asyncIdx=splitCorosWithIdx(callList)
    aLen=0
    aIdx=0
    tasks=[]
    if asyncs:
        for coro in asyncs:
            tasks.append( asyncio.create_task(coro(*args,**kw)) )
        aLen=len(tasks)
    for i,func in enumerate(callList):
        if aIdx<aLen and i==asyncIdx[aIdx]:
            yield await tasks[aIdx]
            aIdx+=1
        else:
            yield func(*args,**kw)

def splitCoros(funclist):
    syncs=[]
    asyncs=[]
    for func in funclist:
        if asyncio.iscoroutinefunction(func):
            asyncs.append(func)
        else:
            syncs.append(func)
    return syncs,asyncs

def splitCorosAsIdx(funclist):
    syncs=[]
    asyncs=[]
    for idx,func in enumerate(funclist):
        if asyncio.iscoroutinefunction(func):
            asyncs.append(idx)
        else:
            syncs.append(idx)
    return syncs,asyncs
    
def splitCorosWithIdx(funclist):
    syncs=[]
    syncIdx=[]
    asyncs=[]
    asyncIdx=[]
    for idx,func in enumerate(funclist):
        if asyncio.iscoroutinefunction(func):
            asyncs.append(func)
            asyncIdx.append(idx)
        else:
            syncs.append(func)
            syncIdx.append(idx)
    return syncs,syncIdx,asyncs,asyncIdx

async def emptyAsyncGen():
    return
    yield


from ..command import CommandCore
from . import EventManager
from . import CommandEvents
# from .command_event import CommandEventsWrapper

import functools

class CommandEventsWrapper:
    """
        为指令添加回调\n
            cmdName 指令名\n
            coreName 指令中枢名，默认为 None，即是使用最后创建的中枢\n
        例：
        ```python
            @Events.onCmd("cmd")
            def cmdExecute(result:ParseResult) -> Any:
                pass

            @Events.onCmd.error("cmd")
            def cmdError(result:ParseResult, err:Exception) -> Any:
                pass
        ```
    """

    def __init__(self,cmdName,coreName=None):
        """
            cmdName 指令名\n
            coreName 指令中枢名，默认为 None，即是使用最后创建的中枢
        """
        core=CommandCore.getCore(coreName)
        cmd=core.cmds.get(cmdName)
        if cmd:
            self.events=cmd.events
        else:
            self.events=CommandEvents(cmdName,core)
    
    def __call__(self,func):
        """
            等同于添加执行回调\n
            回调函数：`(result:ParseResult) -> Any`
        """
        return self._onExecute(func)

    @classmethod
    def execute(cls,cmdName,coreName=None):
        """
            为指令添加执行回调\n
            返回值统一存入解析结果的 `output` 列表中\n
                cmdName 指令名\n
                coreName 指令中枢名，默认为 None，即是使用最后创建的中枢\n
            回调函数：`(result:ParseResult) -> Any`
        """
        return cls(cmdName,coreName)._onExecute

    @classmethod
    def error(cls,cmdName,coreName=None):
        """
            为指令添加报错回调\n
            返回值代替执行回调的结果存入解析结果的 `output` 列表中\n
                cmdName 指令名\n
                coreName 指令中枢名，默认为 None，即是使用最后创建的中枢\n
            回调函数：`(result:ParseResult, err:Exception) -> Any`
        """
        return cls(cmdName,coreName)._onError

    def _onExecute(self,func):
        """
            添加执行回调
        """
        self.events.onExecute(func)
        return func

    def _onError(self,func):
        """
            添加报错回调
        """
        self.events.onError(func)
        return func

def _nameOrFunc(func):
    @functools.wraps(func)
    def wrapper(cls,x=None):
        if x is None or isinstance(x,str):
            return func(cls,x)
        else:
            return func(cls,None)(x)
    return wrapper

class Events:
    """
        提供为各种事件注册回调的装饰器
    """

    @staticmethod
    def getEM(coreName=None) -> EventManager:
        """
            得到指令中枢的事件管理器\n
                coreName 指令中枢名，默认为 None，即是使用最后创建的中枢
        """
        return CommandCore.getCore(coreName).EM

    onCmd=CommandEventsWrapper

    # @staticmethod
    # def onCmd(cmdName,coreName=None) -> CommandEventsWrapper:
    #     return CommandEventsWrapper(cmdName,coreName)

    @classmethod
    def on(cls,name,coreName=None):
        """
            为指定名称的事件添加回调\n
                name 事件名\n
                coreName 指令中枢名，默认为 None，即是使用最后创建的中枢\n
        """
        def decorator(func):
            cls.getEM(coreName).add(name,func)
            return func
        return decorator

    @classmethod
    @_nameOrFunc
    def onNotCmd(cls,coreName=None):
        """
            解析得到非指令时的回调\n
                coreName 指令中枢名，默认为 None，即是使用最后创建的中枢\n
            回调函数：`(result:ParseResult, parser:CommandParser) -> None`
        """
        return cls.on(EventNames.NotCmd,coreName)

    @classmethod
    @_nameOrFunc
    def onUndefinedCmd(cls,coreName=None):
        """
            解析得到未定义指令时的回调\n
                coreName 指令中枢名，默认为 None，即是使用最后创建的中枢\n
            回调函数：`(result:ParseResult, parser:CommandParser) -> None`
        """
        return cls.on(EventNames.UndefinedCmd,coreName)

    @classmethod
    @_nameOrFunc
    def onWrongCmdType(cls,coreName=None):
        """
            解析得到指令，但指令类型（前缀）有误时的回调\n
                coreName 指令中枢名，默认为 None，即是使用最后创建的中枢\n
            回调函数：`(result:ParseResult, parser:CommandParser) -> None`
        """
        return cls.on(EventNames.WrongCmdType,coreName)

    @classmethod
    @_nameOrFunc
    def onBeforeParse(cls,coreName=None):
        """
            每次解析指令参数前的回调\n
                coreName 指令中枢名，默认为 None，即是使用最后创建的中枢\n
            回调函数：`(result:ParseResult, parser:CommandParser) -> None`
        """
        return cls.on(EventNames.BeforeParse,coreName)

    @classmethod
    @_nameOrFunc
    def onAfterParse(cls,coreName=None):
        """
            每次解析指令参数后的回调\n
                coreName 指令中枢名，默认为 None，即是使用最后创建的中枢\n
            回调函数：`(result:ParseResult, parser:CommandParser) -> None`
        """
        return cls.on(EventNames.AfterParse,coreName)
    
    @classmethod
    @_nameOrFunc
    def onExecuteError(cls,coreName=None):
        """
            指令执行出错时的回调\n
            若指令自身无报错回调，且此事件无回调返回 `True`，则继续抛出错误\n
                coreName 指令中枢名，默认为 None，即是使用最后创建的中枢\n
            回调函数：`(result:ParseResult, parser:CommandParser, error:Exception) -> bool`
        """
        return cls.on(EventNames.ExecuteError,coreName)

cmdCallback=Events.onCmd

class EventNames:

    NotCmd="parser-NotCommand"
    UndefinedCmd="parser-UndefinedCmd"
    WrongCmdType="parser-WrongType"
    ExecuteError="parser-ExecuteError"
    BeforeParse="parser-BeforeParse"
    AfterParse="parser-AfterParse"


"""
    指令管理中枢
"""

from ..util.str_similarity import editSimilarity

class CommandCore:
    """
        指令管理中枢\n
        一般会自动创建对象\n
        - CommandCore.getLast() 得到最后创建的指令中枢\n
        - CommandCore.getCore(name) 得到指定名字的指令中枢\n
        也可指定一个名字，手动创建，所有 CommandCore 对象都会存储在 CommandCore.cores 中\n
        配合 with 语句可一定程度上控制指令中枢的生效范围\n
        例如：
        ```python
            with CommandCore(name="admin"):
                pass
        ```
    """
    # corelist=[]
    cores={}
    default="default"
    last=None

    defaultPrefix=(".","。","-","!","！","/")

    @classmethod
    def getLast(cls) -> "CommandCore":
        """
            得到最后创建的指令中枢对象
        """
        if cls.cores:
            if not cls.last:
                cls.last=cls.default
        else:
            CommandCore(name=cls.default)
        return cls.cores[cls.last]

    @classmethod
    def addCore(cls,core,name):
        if name in cls.cores:
            raise ValueError(f"指令管理器 {name} 已存在")
        cls.cores[name]=core
        cls.last=name

    @classmethod
    def getCore(cls,name=None) -> "CommandCore":
        """
            得到指定名字的指令中枢对象\n
            未提供 name 时，同 CommandCore.getLast()\n
                name 指令中枢名
        """
        if name:
            return cls.cores[name] #name 不存在则报错
        else:
            return cls.getLast()

    @property
    def cmdlist(self):
        """
            排序后的指令列表
        """
        self.tryRefresh()
        return self._cmdlist

    def __init__(self,name=None):
        """
            name 中枢名称\n
        """
        self.cmds={}
        self._cmdlist=[]
        self.isRefreshed=False

        from ..event.event_manager import EventManager
        #因为event中也用到CommandCore，所以这里的import出现在奇怪的地方
        self.EM:EventManager=EventManager()

        name=name or CommandCore.default
        self.name=name
        self.oldLast=CommandCore.last

        self.commandPrefix=[*CommandCore.defaultPrefix] #作为中枢内指令的默认前缀
        self.potentialPrefix=set(self.commandPrefix)  #作为中枢内指令可能拥有的潜在前缀，一般认为其包含 commandPrefix

        CommandCore.addCore(self,name)

    def __getitem__(self,key):
        return self.cmds[key] #如果key不存在会报错，只能用来得到已经存在的指令模板

    def __str__(self):
        cmdstr='\n'.join(self.cmdlist)
        return f"指令中枢 {self.name}\n当前记录了{len(self.cmdlist)}个指令名\n{cmdstr}"

    __repr__=__str__

    def __enter__(self):
        CommandCore.last=self.name
        return self

    def __exit__(self,etype,eval,etb):
        CommandCore.last=self.oldLast
        self.tryRefresh()

    def addCommand(self,cmd,name=None):
        if name is None:
            name=cmd.name
        if self.cmds.get(name):
            raise ValueError(f"尝试注册已经存在的指令 {name}")
        self.cmds[name]=cmd
        self.isRefreshed=False #因为添加了新指令，所以可能需要调用refresh方法

    def refresh(self):
        """
            刷新指令列表，并排序\n
        """
        self._cmdlist=list(self.cmds.keys())
        self._cmdlist.sort()
        self.isRefreshed=True
    
    def tryRefresh(self):
        """
            若未刷新，则刷新
        """
        if not self.isRefreshed:
            self.refresh()

    def isMatchPrefix(self,text):
        """
            判断文本是否和前缀列表中的前缀匹配\n
                text 文本\n
            潜在前缀列表见指令中枢对象的 potentialPrefix 属性
        """
        temp=text.lstrip()
        if temp=="":
            return False
        # return temp[0] in self.commandPrefix
        return temp[0] in self.potentialPrefix # 只要命中潜在前缀，都算作是指令


    def getCmdSimilarity(self,text,top=-1,getScore=False,scoreFunc=editSimilarity):
        """
            比较给定文本和列表中指令的相似度\n
                text 文本 如："test"、".test"（指令前缀不会纳入比较）\n
                top 返回列表的长度 默认为-1 若为正，则只返回若干个结果\n
                getScore 是否返回相似度得分 默认为False 若为True，则返回结果中包含相似度\n
                scoreFunc 相似度比较函数 editSimilarity 更多详见 LSparser.util.str_similarity\n
            返回 指令名列表，按和给定文本的相似度降序排列
        """
        perfix=""
        if self.isMatchPrefix(text):
            perfix=text[0]
            text=text[1:]
        similist=[]
        for cn in self.cmdlist:
            score=scoreFunc(text,cn)
            similist.append( (score,cn) )
        similist.sort(reverse=True)
        if getScore:
            resultlist=similist
        else:
            resultlist=[ sl[1] for sl in similist ]
        if top>0:
            resultlist=resultlist[:top]
        return resultlist

if __name__ == "__main__":
    print(CommandCore.last)
    CommandCore.getLast()
    print(CommandCore.last)
    CommandCore(name="extra")
    print(CommandCore.last)
    with CommandCore(name="admin"):
        print(CommandCore.last)
    print(CommandCore.last)
    print(CommandCore.cores)